@@ -15,7 +15,7 @@
  *   You should have received a copy of the GNU General Public License
  *   along with Chiton.  If not, see <https://www.gnu.org/licenses/>.
  *
- *   Copyright 2020-2021 Ed Martin <edman007@edman007.com>
+ *   Copyright 2020-2023 Ed Martin <edman007@edman007.com>
  *
  *   Portions of this file have been borrowed and modified from FFmpeg,
  *    these portions have been origionally licensed under GPL 2.1
@@ -43,6 +43,16 @@
 #ifdef HAVE_OPENCV
 #include <opencv2/core/ocl.hpp>
 #endif
+
+#ifdef HAVE_V4L2
+#include <sys/types.h>
+#include <dirent.h>
+#include <fcntl.h>
+#include <linux/videodev2.h>
+#include <libv4l2.h>
+
+#endif
+
 thread_local std::string last_line;//contains the continuation of any line that doesn't have a \n
 
 CFFUtil gcff_util;//global instance of our util class
@@ -131,6 +141,9 @@ static const struct {
 */
 
 void ffmpeg_log_callback(void * avcl, int level, const char * fmt, va_list vl){
+    if (level > gcff_util.get_ff_log_level()){
+        return;//ignore stuff that is too verbose
+    }
     //compute the level first...
     enum LOG_LEVEL chiton_level;
     if (level <= AV_LOG_FATAL){
@@ -141,17 +154,8 @@ void ffmpeg_log_callback(void * avcl, int level, const char * fmt, va_list vl){
         chiton_level = CH_LOG_INFO;
     } else if (level <= AV_LOG_INFO){
         chiton_level = CH_LOG_INFO;
-    } else if (level <= AV_LOG_VERBOSE){
-        chiton_level = CH_LOG_DEBUG;
-    /* Uncomment to see these messages, they do really cause too much output */
-    /*
-    } else if (level <= AV_LOG_DEBUG){
-        chiton_level = CH_LOG_DEBUG;
-    } else if (level <= AV_LOG_TRACE){
-        chiton_level = CH_LOG_DEBUG;
-    */
     } else {
-        return;//higher level stuff is ignored!
+        chiton_level = CH_LOG_DEBUG;
     }
 
     //format the message
@@ -172,15 +176,13 @@ void ffmpeg_log_callback(void * avcl, int level, const char * fmt, va_list vl){
 }
 
 void CFFUtil::load_ffmpeg(void){
-#ifdef DEBUG
-    av_log_set_level(AV_LOG_DEBUG);
-#else
-    av_log_set_level(AV_LOG_WARNING);
-#endif
     av_log_set_callback(ffmpeg_log_callback);
 
     //init vaapi
     vaapi_ctx = NULL;
+    //we need to always try to load to avoid calling them avcodec_open2() which will cause a deadlock
+    load_vaapi();
+    load_vdpau();
 }
 
 void CFFUtil::load_vaapi(void){
@@ -263,46 +265,61 @@ enum AVPixelFormat get_vdpau_format(AVCodecContext *ctx, const enum AVPixelForma
 }
 
 //Modified list, based on ffmpeg's hwcontext_vaapi.c
-static const AVPixelFormat vaapi_format_map[] = {
-    AV_PIX_FMT_NV12,
-    AV_PIX_FMT_YUV420P,
-    AV_PIX_FMT_YUV422P,
-    AV_PIX_FMT_UYVY422,
-    AV_PIX_FMT_YUYV422,
+static const struct {
+    unsigned int va;//vaapi fourcc
+    AVPixelFormat ff;//ffmpeg
+} vaapi_format_map[] = {
+    {VA_FOURCC_NV12, AV_PIX_FMT_NV12},
+#ifdef VA_FOURCC_I420
+    {VA_FOURCC_I420,  AV_PIX_FMT_YUV420P},
+#endif
+    {VA_FOURCC_YV12,  AV_PIX_FMT_YUV420P},
+    {VA_FOURCC_IYUV,  AV_PIX_FMT_YUV420P},
+    {VA_FOURCC_422H,  AV_PIX_FMT_YUV422P},
+#ifdef VA_FOURCC_YV16
+    {VA_FOURCC_YV16,  AV_PIX_FMT_YUV422P},
+#endif
+    {VA_FOURCC_UYVY,  AV_PIX_FMT_UYVY422},
+    {VA_FOURCC_YUY2,  AV_PIX_FMT_YUYV422},
 #ifdef VA_FOURCC_Y210
-#ifdef AV_PIX_FMT_Y210
-    AV_PIX_FMT_Y210,
+    {VA_FOURCC_Y210,  AV_PIX_FMT_Y210},
 #endif
+#if defined(VA_FOURCC_Y212) && defined(AV_PIX_FMT_Y212)
+    {VA_FOURCC_Y212,  AV_PIX_FMT_Y212},
 #endif
-    AV_PIX_FMT_YUV411P,
-    AV_PIX_FMT_YUV440P,
-    AV_PIX_FMT_YUV444P,
-    AV_PIX_FMT_GRAY8,
+    {VA_FOURCC_411P,  AV_PIX_FMT_YUV411P},
+    {VA_FOURCC_422V,  AV_PIX_FMT_YUV440P},
+    {VA_FOURCC_444P,  AV_PIX_FMT_YUV444P},
+#if defined(VA_FOURCC_XYUV) && defined(AV_PIX_FMT_VUYX)
+    {VA_FOURCC_XYUV,  AV_PIX_FMT_VUYX},
+#endif
+    {VA_FOURCC_Y800,  AV_PIX_FMT_GRAY8},
 #ifdef VA_FOURCC_P010
-#ifdef AV_PIX_FMT_P010
-    AV_PIX_FMT_P010,
+    {VA_FOURCC_P010, AV_PIX_FMT_P010},
 #endif
+#if defined(VA_FOURCC_P012) && defined(AV_PIX_FMT_P012)
+    {VA_FOURCC_P012, AV_PIX_FMT_P012},
 #endif
-    AV_PIX_FMT_BGRA,
-    AV_PIX_FMT_BGR0,
-    AV_PIX_FMT_RGBA,
-    AV_PIX_FMT_RGB0,
+    {VA_FOURCC_BGRA,   AV_PIX_FMT_BGRA},
+    {VA_FOURCC_BGRX,   AV_PIX_FMT_BGR0},
+    {VA_FOURCC_RGBA,   AV_PIX_FMT_RGBA},
+    {VA_FOURCC_RGBX,   AV_PIX_FMT_RGB0},
 #ifdef VA_FOURCC_ABGR
-#ifdef AV_PIX_FMT_ABGR
-    AV_PIX_FMT_ABGR,
+    {VA_FOURCC_ABGR,   AV_PIX_FMT_ABGR},
+    {VA_FOURCC_XBGR,   AV_PIX_FMT_0BGR},
 #endif
-#ifdef AV_PIX_FMT_0BGR
-    AV_PIX_FMT_0BGR,
+    {VA_FOURCC_ARGB,   AV_PIX_FMT_ARGB},
+    {VA_FOURCC_XRGB,   AV_PIX_FMT_0RGB},
+#if defined(VA_FOURCC_X2R10G10B10) && defined(AV_PIX_FMT_X2RGB10)
+    {VA_FOURCC_X2R10G10B10, AV_PIX_FMT_X2RGB10},
 #endif
+#if defined(VA_FOURCC_Y410) && defined(AV_PIX_FMT_XV30)
+    {VA_FOURCC_Y410, AV_PIX_FMT_XV30},
 #endif
-    AV_PIX_FMT_ARGB,
-    AV_PIX_FMT_0RGB,
-#ifdef VA_FOURCC_X2R10G10B10
-#ifdef AV_PIX_FMT_X2RGB10
-    AV_PIX_FMT_X2RGB10,
+#if defined(VA_FOURCC_Y412) && defined(AV_PIX_FMT_XV36)
+    {VA_FOURCC_Y412,  AV_PIX_FMT_XV36},
 #endif
-#endif
-    AV_PIX_FMT_NONE
+    {0, AV_PIX_FMT_NONE}
 };
 
 
@@ -343,7 +360,7 @@ bool CFFUtil::sw_format_is_hw_compatable(const enum AVPixelFormat pix_fmt){
     return false;
 }
 
-std::string CFFUtil::get_sw_hw_format_list(Config &cfg){
+std::string CFFUtil::get_sw_hw_format_list(Config &cfg, const AVFrame *frame, AVCodecID codec_id, int codec_profile){
     load_vaapi();
     if (!vaapi_ctx){
         return "";
@@ -354,12 +371,55 @@ std::string CFFUtil::get_sw_hw_format_list(Config &cfg){
             return std::string(cfg_fmt);
         }
     }
-    AVHWFramesConstraints* c = av_hwdevice_get_hwframe_constraints(vaapi_ctx, NULL);
+    std::stringstream ss;
+    void* hw_config = NULL;
+#ifdef HAVE_VAAPI
+    if (frame && frame->format == AV_PIX_FMT_VAAPI && frame->hw_frames_ctx){
+        AVVAAPIDeviceContext *hwctx = get_vaapi_ctx_from_frames(frame->hw_frames_ctx);;
+        AVHWFramesContext *frames = reinterpret_cast<AVHWFramesContext*>(frame->hw_frames_ctx->data);
+        hw_config = av_hwdevice_hwconfig_alloc(frames->device_ref);
+        if (hw_config){
+            AVVAAPIHWConfig *va_hw_config = static_cast<AVVAAPIHWConfig*>(hw_config);
+
+            //the stupid defaults
+            VAEntrypoint entry = VAEntrypointVLD;//this is correct for FFMPeg decode, but does NOT work with VAProfileNone
+            VAProfile prof = VAProfileNone;
+
+            if (codec_id != AV_CODEC_ID_NONE){
+                prof = get_va_profile(hwctx, codec_id, codec_profile);
+            }
+            if (prof == VAProfileNone){
+                entry = VAEntrypointVideoProc;//if we didn't get a profile, we set to this to be more generic
+            }
+            VAStatus cfg_stat = vaCreateConfig(hwctx->display, prof, entry, NULL, 0, &va_hw_config->config_id);
+            if (cfg_stat != VA_STATUS_SUCCESS){
+                LWARN("vaCreateConfig Failed");
+                av_free(hw_config);
+                hw_config = NULL;
+            }
+        }
+
+    }
+#endif
+    AVHWFramesConstraints* c = av_hwdevice_get_hwframe_constraints(vaapi_ctx, hw_config);
+    if (hw_config){
+        //free the hw config
+#ifdef HAVE_VAAPI
+        if (frame && frame->format == AV_PIX_FMT_VAAPI){
+            AVVAAPIDeviceContext *hwctx = get_vaapi_ctx_from_frames(frame->hw_frames_ctx);;
+            if (hwctx){
+                AVVAAPIHWConfig *va_hw_config = static_cast<AVVAAPIHWConfig*>(hw_config);
+                vaDestroyConfig(hwctx->display, va_hw_config->config_id);
+            }
+        }
+#endif
+        av_free(hw_config);
+    }
     if (!c){
         return "";
     }
 
-    std::stringstream ss;
+
     const enum AVPixelFormat *v;
     for (v = c->valid_sw_formats; *v != AV_PIX_FMT_NONE; v++){
         if (v != c->valid_sw_formats){
@@ -386,43 +446,11 @@ bool CFFUtil::have_vaapi(AVCodecID codec_id, int codec_profile, int width, int h
     if (!codec_desc) {
         return NULL;
     }
-    VAProfile profile, *profile_list = NULL;
-    int profile_count;
-    profile_count = vaMaxNumProfiles(hwctx->display);
-    profile_list = new VAProfile[profile_count];
-    if (!profile_list){
-        return NULL;
-    }
-
-    VAStatus vas;
-    vas = vaQueryConfigProfiles(hwctx->display, profile_list, &profile_count);
-    if (vas != VA_STATUS_SUCCESS) {
-        delete[] profile_list;
-        return NULL;
-    }
-
-    profile = VAProfileNone;
-
-    //search all known codecs to see if there is a matching profile
-    for (unsigned int i = 0; i < FF_ARRAY_ELEMS(vaapi_profile_map); i++) {
-        if (codec_id != vaapi_profile_map[i].codec_id){
-            continue;
-        }
-        for (int j = 0; j < profile_count; j++) {
-            if (vaapi_profile_map[i].va_profile == profile_list[j] &&
-                vaapi_profile_map[i].codec_profile == codec_profile) {
-                profile = profile_list[j];
-                break;//exact found, we are done
-            } else if (vaapi_profile_map[i].va_profile == profile_list[j]){
-                profile = profile_list[j];//in exact found
-            }
-        }
-    }
-    delete[] profile_list;
+    VAProfile profile = get_va_profile(hwctx, codec_id, codec_profile);
 
     //if we couldn't find a matching profile we bail
     if (profile == VAProfileNone){
-        LINFO("VA-API does not support this codec, no profile found");
+        LINFO("VA-API does not support this codec (" + std::string(avcodec_get_name(codec_id)) + "/" + avcodec_profile_name(codec_id, codec_profile) + "), no profile found");
         return NULL;
     }
 #endif
@@ -514,6 +542,7 @@ AVBufferRef *CFFUtil::get_vdpau_ctx(AVCodecID codec_id, int codec_profile, int w
 CFFUtil::CFFUtil(void){
     vaapi_failed = false;
     vdpau_failed = false;
+    max_ffmpeg_log_level = AV_LOG_VERBOSE;
 }
 
 CFFUtil::~CFFUtil(void){
@@ -593,64 +622,13 @@ AVBufferRef *CFFUtil::get_opencl_ctx(AVCodecID codec_id, int codec_profile, int
 }
 
 bool CFFUtil::have_opencl(AVCodecID codec_id, int codec_profile, int width, int height){
-    load_vaapi();
-    if (!vaapi_ctx){
+    load_opencl();
+    if (!opencl_ctx){
         return NULL;
     }
 
-    //this section is only used to see if VAAPI supports our codec, without VAAPI we can still ask libavcodec to do it, but we won't known
-    //if it's going to work, this will result in is just failing if we try it and it's not supported
-#ifdef HAVE_VAAPI
-    AVVAAPIDeviceContext* hwctx = reinterpret_cast<AVVAAPIDeviceContext*>((reinterpret_cast<AVHWDeviceContext*>(vaapi_ctx->data))->hwctx);
-    //query VAAPI profiles for this codec
-    const AVCodecDescriptor *codec_desc;
-    codec_desc = avcodec_descriptor_get(codec_id);
-    if (!codec_desc) {
-        return NULL;
-    }
-    VAProfile profile, *profile_list = NULL;
-    int profile_count;
-    profile_count = vaMaxNumProfiles(hwctx->display);
-    profile_list = new VAProfile[profile_count];
-    if (!profile_list){
-        return NULL;
-    }
-
-    VAStatus vas;
-    vas = vaQueryConfigProfiles(hwctx->display, profile_list, &profile_count);
-    if (vas != VA_STATUS_SUCCESS) {
-        delete[] profile_list;
-        return NULL;
-    }
-
-    profile = VAProfileNone;
-
-    //search all known codecs to see if there is a matching profile
-    for (unsigned int i = 0; i < FF_ARRAY_ELEMS(vaapi_profile_map); i++) {
-        if (codec_id != vaapi_profile_map[i].codec_id){
-            continue;
-        }
-        for (int j = 0; j < profile_count; j++) {
-            if (vaapi_profile_map[i].va_profile == profile_list[j] &&
-                vaapi_profile_map[i].codec_profile == codec_profile) {
-                profile = profile_list[j];
-                break;//exact found, we are done
-            } else if (vaapi_profile_map[i].va_profile == profile_list[j]){
-                profile = profile_list[j];//in exact found
-            }
-        }
-    }
-    delete[] profile_list;
-
-    //if we couldn't find a matching profile we bail
-    if (profile == VAProfileNone){
-        LINFO("VA-API does not support this codec, no profile found");
-        return NULL;
-    }
-#endif
-
     bool found = false;
-    AVHWFramesConstraints* c = av_hwdevice_get_hwframe_constraints(vaapi_ctx, NULL);
+    AVHWFramesConstraints* c = av_hwdevice_get_hwframe_constraints(opencl_ctx, NULL);
     if (!c){
         return NULL;
     }
@@ -762,3 +740,222 @@ AVOpenCLDeviceContext *get_opencl_ctx_from_frames(AVBufferRef *buf){
     return get_opencl_ctx_from_device(frames->device_ref);
 }
 #endif
+
+
+bool CFFUtil::have_v4l2(AVCodecID codec_id, int codec_profile, int width, int height){
+#ifdef HAVE_V4L2
+    //we need to check all devices
+    DIR *dirp = opendir("/dev");
+    if (!dirp){
+        LWARN("Can't open /dev");
+        return false;
+    }
+
+    bool ret = false;
+    struct dirent *entry;
+    while (!ret && (entry = readdir(dirp))){
+        std::string dev_name = std::string(entry->d_name);
+        if (dev_name.find("video") != 0){
+            continue;
+        }
+        int fd = v4l2_open(std::string("/dev/" + dev_name).c_str(), O_RDWR | O_NONBLOCK, 0);
+        if (fd < 0){
+            LWARN("Cannot open /dev/" + dev_name + " (" + std::to_string(errno) + ")");
+            continue;
+        }
+        ret = have_v4l2_dev(fd, codec_id, codec_profile, width, height);
+        if (v4l2_close(fd)){
+            LWARN("close() failed (" + std::to_string(errno) + ")");
+        }
+    }
+    closedir(dirp);
+    LDEBUG("Completed dev check");
+    return ret;
+    //query pixel formats first
+#else
+    return false;//no v4l2 enabled
+#endif
+}
+
+#ifdef HAVE_V4L2
+bool CFFUtil::have_v4l2_dev(int dev, AVCodecID codec_id, int codec_profile, int width, int height){
+    bool ret = false;
+
+    //check if it supports M2M
+    struct v4l2_capability caps;
+    if (!vioctl(dev, VIDIOC_QUERYCAP, &caps)){
+        return false;
+    }
+
+    LDEBUG("Found V4L2 device " + std::string(reinterpret_cast<char*>(caps.card)) + "[" + std::string(reinterpret_cast<char*>(caps.bus_info)) + "]");
+
+    bool splane = caps.device_caps & V4L2_CAP_VIDEO_M2M;
+    bool mplane = caps.device_caps & V4L2_CAP_VIDEO_M2M_MPLANE;
+    if (!splane && !mplane){
+        LDEBUG("...Does not support M2M");
+        return false;
+    }
+
+    struct v4l2_fmtdesc pixfmt;
+    for (int pix_idx = 0;; pix_idx++){
+        pixfmt.index = pix_idx;
+        pixfmt.type = splane ? V4L2_BUF_TYPE_VIDEO_OUTPUT : V4L2_BUF_TYPE_VIDEO_OUTPUT_MPLANE;
+        if (!vioctl(dev, VIDIOC_ENUM_FMT, &pixfmt)){
+            if (splane && mplane){
+                //not sure if anyone supports both on the same driver, but we do!
+                //this repeates the loop for mplane
+                splane = false;
+                pix_idx = -1;
+                continue;
+            }
+            break;//hit the end
+        }
+
+        if (pixfmt.flags & V4L2_FMT_FLAG_EMULATED){
+            continue;//we don't care about SW formats
+        }
+
+        if (!is_v4l2_hw_codec(codec_id, pixfmt.pixelformat)){
+            continue;//now the requested codec, irrelevant
+        }
+        struct v4l2_frmsizeenum frsize;
+        for (int fr_idx = 0; ; fr_idx++){
+            frsize.index = fr_idx;
+            frsize.pixel_format = pixfmt.pixelformat;
+            if (!vioctl(dev, VIDIOC_ENUM_FRAMESIZES, &frsize)){
+                break;
+            }
+            if (frsize.type == V4L2_FRMSIZE_TYPE_DISCRETE){
+                if (width == static_cast<int>(frsize.discrete.width) && height == static_cast<int>(frsize.discrete.height)){
+                    return true;
+                }
+                break;//DISCRETE is only one
+            } else if (frsize.type == V4L2_FRMSIZE_TYPE_STEPWISE || frsize.type == V4L2_FRMSIZE_TYPE_CONTINUOUS) {
+                if (width >= static_cast<int>(frsize.stepwise.min_width) &&
+                    width <= static_cast<int>(frsize.stepwise.max_width) &&
+                    width % frsize.stepwise.step_width == 0 &&
+                    height >= static_cast<int>(frsize.stepwise.min_height) &&
+                    height <= static_cast<int>(frsize.stepwise.max_height) &&
+                    height % frsize.stepwise.step_height == 0){
+                    return true;
+                }
+            } else {
+                LWARN("Unknown V4L2 Type - " + std::to_string(frsize.type));
+                return false;
+            }
+        }
+    }
+    return ret;
+}
+#endif
+
+#ifdef HAVE_V4L2
+bool CFFUtil::vioctl(int fh, int request, void *arg){
+    int r;
+
+    do {
+        r = v4l2_ioctl(fh, request, arg);
+    } while (r == -1 && ((errno == EINTR) || (errno == EAGAIN)));
+
+    if (r == -1){
+        return false;
+    }
+    return true;
+}
+#endif
+
+#ifdef HAVE_V4L2
+bool CFFUtil::is_v4l2_hw_codec(const AVCodecID av_codec, const uint32_t v4l2_pix_fmt){
+    //i hate these things
+    //we only use this to determine if FFMpeg will probably succeed, so it doesn't have to be 100% accurate
+    switch (av_codec){
+    case AV_CODEC_ID_MPEG1VIDEO:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_MPEG1;
+    case AV_CODEC_ID_MPEG2VIDEO:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_MPEG2;
+    case AV_CODEC_ID_H263:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_H263;
+    case AV_CODEC_ID_H264:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_H264 ||
+            v4l2_pix_fmt == V4L2_PIX_FMT_H264_NO_SC ||
+            v4l2_pix_fmt == V4L2_PIX_FMT_H264_MVC;
+    case AV_CODEC_ID_MPEG4:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_MPEG4 ||
+            v4l2_pix_fmt == V4L2_PIX_FMT_XVID;
+    case AV_CODEC_ID_VC1:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_VC1_ANNEX_G ||
+            v4l2_pix_fmt == V4L2_PIX_FMT_VC1_ANNEX_L;
+    case AV_CODEC_ID_HEVC:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_HEVC;
+    case AV_CODEC_ID_VP8:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_VP8;
+    case AV_CODEC_ID_VP9:
+        return v4l2_pix_fmt == V4L2_PIX_FMT_VP9;
+    default:
+        return false;
+    }
+}
+#endif
+
+
+#ifdef HAVE_VAAPI
+AVPixelFormat CFFUtil::get_pix_fmt_from_va(const VAImageFormat &fmt){
+    for (unsigned int i = 0; i < (sizeof(vaapi_format_map)/sizeof(vaapi_format_map[0])); i++){
+        if (fmt.fourcc == vaapi_format_map[i].va){
+            return vaapi_format_map[i].ff;
+        }
+    }
+    return AV_PIX_FMT_NONE;
+}
+#endif
+
+#ifdef HAVE_VAAPI
+VAProfile CFFUtil::get_va_profile(AVVAAPIDeviceContext* hwctx, AVCodecID codec_id, int codec_profile){
+    VAProfile profile, *profile_list;
+    int profile_count;
+    VAStatus vas;
+    profile = VAProfileNone;
+
+    profile_count = vaMaxNumProfiles(hwctx->display);
+    profile_list = new VAProfile[profile_count];
+    if (!profile_list){
+        return VAProfileNone;
+    }
+
+    vas = vaQueryConfigProfiles(hwctx->display, profile_list, &profile_count);
+    if (vas != VA_STATUS_SUCCESS) {
+        delete[] profile_list;
+        return VAProfileNone;
+    }
+
+    //search all known codecs to see if there is a matching profile
+    for (unsigned int i = 0; i < FF_ARRAY_ELEMS(vaapi_profile_map); i++) {
+        if (codec_id != vaapi_profile_map[i].codec_id){
+            continue;
+        }
+        for (int j = 0; j < profile_count; j++) {
+            if (vaapi_profile_map[i].va_profile == profile_list[j] &&
+                vaapi_profile_map[i].codec_profile == codec_profile) {
+                profile = profile_list[j];
+                break;//exact found, we are done
+            } else if (vaapi_profile_map[i].va_profile == profile_list[j]){
+                profile = profile_list[j];//inexact found
+            }
+        }
+    }
+    delete[] profile_list;
+    return profile;
+}
+#endif
+
+int CFFUtil::get_ff_log_level(void) const{
+    return max_ffmpeg_log_level;
+}
+
+void CFFUtil::set_ff_log_level(int lvl){
+    if (lvl < AV_LOG_QUIET || lvl > AV_LOG_TRACE){
+        LINFO("ffmpeg-log-level appears unreasonable, typically between " + std::to_string(AV_LOG_QUIET) + " and " + std::to_string(AV_LOG_TRACE));
+    }
+    max_ffmpeg_log_level = lvl;
+    av_log_set_level(lvl);
+}
