@@ -17,7 +17,7 @@ class ScopeTest extends PHPUnit_Smarty
 {
     public function setUp(): void
     {
-        $this->setUpSmarty(dirname(__FILE__));
+        $this->setUpSmarty(__DIR__);
         $this->smarty->addPluginsDir("../../__shared/PHPunitplugins/");
         $this->smarty->addTemplateDir("../../__shared/templates/");
         $this->smarty->addTemplateDir("./templates_tmp");
