@@ -15,33 +15,41 @@
  *   You should have received a copy of the GNU General Public License
  *   along with Chiton.  If not, see <https://www.gnu.org/licenses/>.
  *
- *   Copyright 2020-2022 Ed Martin <edman007@edman007.com>
+ *   Copyright 2020-2023 Ed Martin <edman007@edman007.com>
  *
  **************************************************************************
  */
 #include "file_manager.hpp"
 
+#include "util.hpp"
+#include "system_controller.hpp"
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
 #include <sys/statvfs.h>
 #include <dirent.h>
-#include "util.hpp"
 
 //static globals
 std::mutex FileManager::cleanup_mtx;//lock when cleanup is in progress, locks just the clean_disk() to prevent iterating over the same database results twice
 std::atomic<long long> FileManager::reserved_bytes(0);
 
-FileManager::FileManager(Database &db, Config &cfg) : db(db), cfg(cfg) {
+FileManager::FileManager(SystemController &sys) : FileManager(sys, sys.get_sys_cfg()) {
+    //no-op, calls detailed version
+}
+
+FileManager::FileManager(SystemController &sys, Config &cfg) : sys(sys), db(sys.get_db()), cfg(cfg) {
     bytes_per_segment = 1024*1024;//1M is our default guess
     min_free_bytes = -1;
+}
 
+void FileManager::init(void){
     //we correct this if it's wrong now, so we can assume it's always right
     const std::string &ext = cfg.get_value("output-extension");
     if (ext != ".ts" && ext != ".mp4"){
         LWARN("output-extension MUST be .ts or .mp4, using .mp4");
         cfg.set_value("output-extension", ".mp4");
     }
+
 }
 
 std::string FileManager::get_next_path(long int &file_id, int camera, const struct timeval &start_time, bool extend_file /* = false */){
@@ -194,7 +202,7 @@ void FileManager::delete_broken_segments(void){
     std::string future_time = std::to_string(Util::pack_time(curtime));
 
     std::string sql = "SELECT v.name, v.path, c.value, v.extension, v.camera, v.starttime FROM videos AS v LEFT JOIN config AS c ON v.camera=c.camera AND c.name = 'output-dir' WHERE "
-        "v.endtime < v.starttime OR v.starttime > " + future_time;
+        "v.endtime IS NULL OR v.endtime < v.starttime OR v.starttime > " + future_time;
     DatabaseResult* res = db.query(sql);
     if (res && res->num_rows() > 0){
         LWARN("Found " + std::to_string(res->num_rows()) + " broken  segments, deleting them");
