@@ -42,12 +42,14 @@ bool StreamUnwrap::close(void){
     //the reorder queue should be freed
     for (auto pkt : reorder_queue){
         unref_frame(pkt);
+        av_packet_free(&pkt);
     }
     reorder_queue.clear();
     reorder_len = 0;
 
     for (auto &pkt : decoded_packets){
-        av_packet_unref(&pkt);
+        av_packet_unref(pkt);
+        av_packet_free(&pkt);
     }
     for (auto &frame : decoded_video_frames){
         av_frame_free(&frame);
@@ -156,6 +158,18 @@ bool StreamUnwrap::alloc_decode_context(unsigned int stream){
         return false;
     }
 
+    AVCodecParameters *stream_codecpar = input_format_context->streams[stream]->codecpar;
+    bool using_v4l2 = false;
+    //check if we need to do V4L2
+    if ((cfg.get_value("video-decode-method") == "auto" &&
+        !gcff_util.have_vaapi(stream_codecpar->codec_id, stream_codecpar->profile, stream_codecpar->height, stream_codecpar->width) &&
+         !gcff_util.have_vdpau(stream_codecpar->codec_id, stream_codecpar->profile, stream_codecpar->height, stream_codecpar->width)) ||
+        cfg.get_value("video-decode-method") == "v4l2"){
+        LDEBUG("Attempting V4L2");
+        //use v4l2
+        using_v4l2 = gcff_util.have_v4l2(stream_codecpar->codec_id, stream_codecpar->profile, stream_codecpar->height, stream_codecpar->width);
+    }
+
     //locate a decoder for the codec
     input_codec = avcodec_find_decoder(input_format_context->streams[stream]->codecpar->codec_id);
     if (!input_codec){
@@ -163,6 +177,23 @@ bool StreamUnwrap::alloc_decode_context(unsigned int stream){
         return false;
     }
 
+    if (using_v4l2){
+        //check if this codec has a v4l2 version
+        std::string v4l2_name = std::string(input_codec->name) + std::string("_v4l2m2m");
+        const AVCodec *v4l_codec = avcodec_find_decoder_by_name(v4l2_name.c_str());
+        if (v4l_codec){
+            LINFO("Using ffmpeg " + std::string(v4l_codec->long_name) + " for decoding");
+            input_codec = v4l_codec;
+        } else {
+            using_v4l2 = false;
+            LINFO("You seem to have V4L2 support for " + std::string(input_codec->long_name) + " but ffmpeg was not compiled with the associated support");
+        }
+    }
+
+    if (!using_v4l2 && cfg.get_value("video-decode-method") == "v4l2"){
+        LWARN("V4L2 decoding was requested but is not available, falling back to SW");
+    }
+
     /* Allocate a new decoding context. */
     avctx = avcodec_alloc_context3(input_codec);
     if (!avctx) {
@@ -179,7 +210,7 @@ bool StreamUnwrap::alloc_decode_context(unsigned int stream){
 
     if (avctx->codec_type == AVMEDIA_TYPE_VIDEO){
         //connect decoder
-        if (!avctx->hw_device_ctx &&
+        if (!using_v4l2 && !avctx->hw_device_ctx &&
             (cfg.get_value("video-decode-method") == "auto" || cfg.get_value("video-decode-method") == "vaapi")){
             avctx->hw_device_ctx  = gcff_util.get_vaapi_ctx(avctx->codec_id, avctx->profile, avctx->height, avctx->width);
             if (avctx->hw_device_ctx){
@@ -187,7 +218,7 @@ bool StreamUnwrap::alloc_decode_context(unsigned int stream){
                 LINFO("Using VA-API for decoding");
             }
         }
-        if (!avctx->hw_device_ctx &&
+        if (!using_v4l2 && !avctx->hw_device_ctx &&
             (cfg.get_value("video-decode-method") == "auto" || cfg.get_value("video-decode-method") == "vdpau")){
             avctx->hw_device_ctx  = gcff_util.get_vdpau_ctx(avctx->codec_id, avctx->profile, avctx->height, avctx->width);
             if (avctx->hw_device_ctx){
@@ -199,7 +230,9 @@ bool StreamUnwrap::alloc_decode_context(unsigned int stream){
         if (!avctx->hw_device_ctx){
             avctx->opaque = this;
             avctx->get_format = get_frame_format;//do any format, but prefer VAAPI compatible formats
-            LINFO("Using SW Decoding");
+            if (!using_v4l2){
+                LINFO("Using SW Decoding");
+            }
         }
     }
 
@@ -219,9 +252,10 @@ bool StreamUnwrap::alloc_decode_context(unsigned int stream){
 
 }
 
-bool StreamUnwrap::get_next_frame(AVPacket &packet){
+bool StreamUnwrap::get_next_frame(AVPacket *packet){
     if (!decoded_packets.empty()){
-        av_packet_move_ref(&packet, &decoded_packets.front());
+        av_packet_move_ref(packet, decoded_packets.front());
+        av_packet_free(&decoded_packets.front());
         decoded_packets.pop_front();
         return true;
     }
@@ -229,18 +263,19 @@ bool StreamUnwrap::get_next_frame(AVPacket &packet){
 
 }
 
-bool StreamUnwrap::get_next_packet(AVPacket &packet){
+bool StreamUnwrap::get_next_packet(AVPacket *packet){
     read_frame();
     if (reorder_queue.size() > 0){
-        av_packet_move_ref(&packet, &reorder_queue.front());
+        av_packet_move_ref(packet, reorder_queue.front());
+        av_packet_free(&reorder_queue.front());
         reorder_queue.pop_front();
 
         //fixup the duration
-        if (packet.duration == 0 || packet.duration == AV_NOPTS_VALUE){
-            packet.duration = 0;
+        if (packet->duration == 0 || packet->duration == AV_NOPTS_VALUE){
+            packet->duration = 0;
             for (const auto &next : reorder_queue){
-                if (next.stream_index == packet.stream_index){
-                    packet.duration = next.pts - packet.pts;
+                if (next->stream_index == packet->stream_index){
+                    packet->duration = next->pts - packet->pts;
                     break;
                 }
             }
@@ -253,8 +288,8 @@ bool StreamUnwrap::get_next_packet(AVPacket &packet){
     }
     
 }
-void StreamUnwrap::unref_frame(AVPacket &packet){
-    av_packet_unref(&packet);
+void StreamUnwrap::unref_frame(AVPacket *packet){
+    av_packet_unref(packet);
 }
 
 void StreamUnwrap::dump_options(AVDictionary* dict){
@@ -280,40 +315,40 @@ void StreamUnwrap::sort_reorder_queue(void){
     }
     auto av_old_back = reorder_queue.end();
     std::advance(av_old_back, -2);
-    AVPacket &old_back = *av_old_back;
-    long back_dts_in_old_back = av_rescale_q_rnd(reorder_queue.back().dts,
-                                                 input_format_context->streams[reorder_queue.back().stream_index]->time_base,
-                                                 input_format_context->streams[old_back.stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
+    AVPacket *old_back = *av_old_back;
+    long back_dts_in_old_back = av_rescale_q_rnd(reorder_queue.back()->dts,
+                                                 input_format_context->streams[reorder_queue.back()->stream_index]->time_base,
+                                                 input_format_context->streams[old_back->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
     
     
-    if (old_back.dts > back_dts_in_old_back){
+    if (old_back->dts > back_dts_in_old_back){
         //need to walk the queue and insert it in the right spot...
-        bool back_is_video = input_format_context->streams[reorder_queue.back().stream_index]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO;
+        bool back_is_video = input_format_context->streams[reorder_queue.back()->stream_index]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO;
         //FIXME: this loop should be optimized to run back to front, as that's the short direction
-        for (std::list<AVPacket>::iterator itr = reorder_queue.begin(); itr != reorder_queue.end() ; itr++){
-            long back_dts_in_itr = av_rescale_q_rnd(reorder_queue.back().dts,
-                                                    input_format_context->streams[reorder_queue.back().stream_index]->time_base,
-                                                    input_format_context->streams[itr->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
-            if ((itr->dts > reorder_queue.back().dts && itr->stream_index == reorder_queue.back().stream_index) || //same stream and dts is out of order
-                (itr->stream_index != reorder_queue.back().stream_index && itr->dts > back_dts_in_itr) || // different stream and out of order
-                (itr->stream_index != reorder_queue.back().stream_index && itr->dts == back_dts_in_itr && back_is_video)){ //different stream and video and same time
+        for (std::list<AVPacket*>::iterator itr = reorder_queue.begin(); itr != reorder_queue.end() ; itr++){
+            long back_dts_in_itr = av_rescale_q_rnd(reorder_queue.back()->dts,
+                                                    input_format_context->streams[reorder_queue.back()->stream_index]->time_base,
+                                                    input_format_context->streams[(*itr)->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
+            if (((*itr)->dts > reorder_queue.back()->dts && (*itr)->stream_index == reorder_queue.back()->stream_index) || //same stream and dts is out of order
+                ((*itr)->stream_index != reorder_queue.back()->stream_index && (*itr)->dts > back_dts_in_itr) || // different stream and out of order
+                ((*itr)->stream_index != reorder_queue.back()->stream_index && (*itr)->dts == back_dts_in_itr && back_is_video)){ //different stream and video and same time
                 if (itr == reorder_queue.begin() && reorder_queue.size() == reorder_len){
                     //itr is the front here
                     //estimate what it would take
-                    long dts_itr_to_front = itr->dts - back_dts_in_itr;//this is how far ahead the packet was from the queue
+                    long dts_itr_to_front = (*itr)->dts - back_dts_in_itr;//this is how far ahead the packet was from the queue
                     auto av_back = reorder_queue.end();
                     std::advance(av_back, -2);
-                    AVPacket &back = *av_back;
+                    AVPacket *back = *av_back;
                     
-                    AVPacket &front = reorder_queue.front();
-                    long back_dts_in_itr = av_rescale_q_rnd(back.dts,
-                                                            input_format_context->streams[back.stream_index]->time_base,
-                                                            input_format_context->streams[itr->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
-                    long front_dts_in_itr = av_rescale_q_rnd(front.dts,
-                                                             input_format_context->streams[front.stream_index]->time_base,
-                                                             input_format_context->streams[itr->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
-                    double que_lent = av_q2d(input_format_context->streams[itr->stream_index]->time_base) * ( back_dts_in_itr - front_dts_in_itr);
-                    double target_addational_lent = av_q2d(input_format_context->streams[itr->stream_index]->time_base) * ( dts_itr_to_front );
+                    AVPacket *front = reorder_queue.front();
+                    long back_dts_in_itr = av_rescale_q_rnd(back->dts,
+                                                            input_format_context->streams[back->stream_index]->time_base,
+                                                            input_format_context->streams[(*itr)->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
+                    long front_dts_in_itr = av_rescale_q_rnd(front->dts,
+                                                             input_format_context->streams[front->stream_index]->time_base,
+                                                             input_format_context->streams[(*itr)->stream_index]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
+                    double que_lent = av_q2d(input_format_context->streams[(*itr)->stream_index]->time_base) * ( back_dts_in_itr - front_dts_in_itr);
+                    double target_addational_lent = av_q2d(input_format_context->streams[(*itr)->stream_index]->time_base) * ( dts_itr_to_front );
                     double target_change = (target_addational_lent/que_lent) + 1.0;
                     LINFO("Reorder queue is short by " + std::to_string(target_addational_lent) + "s, recommend making reorder-queue-len at least " +
                           std::to_string((int)(target_change * reorder_queue.size())) + " for camera " /*+ cfg.get_value("camera-id") */);
@@ -329,11 +364,12 @@ void StreamUnwrap::sort_reorder_queue(void){
 }
 
 bool StreamUnwrap::read_frame(void){
-    reorder_queue.emplace_back();
+    reorder_queue.emplace_back(av_packet_alloc());
     struct timeval start_recv_time;
     Util::get_videotime(start_recv_time);
 
-    if (av_read_frame(input_format_context, &reorder_queue.back())){
+    if (av_read_frame(input_format_context, reorder_queue.back())){
+        av_packet_free(&reorder_queue.back());
         reorder_queue.pop_back();
         LINFO("av_read_frame() failed or hit EOF");
         return false;
@@ -344,11 +380,11 @@ bool StreamUnwrap::read_frame(void){
 
 
     //fixup this data...maybe limit it to the first one?
-    if (reorder_queue.back().dts == AV_NOPTS_VALUE){
-        reorder_queue.back().dts = 0;
+    if (reorder_queue.back()->dts == AV_NOPTS_VALUE){
+        reorder_queue.back()->dts = 0;
     }
-    if (reorder_queue.back().pts == AV_NOPTS_VALUE){
-        reorder_queue.back().pts = 0;
+    if (reorder_queue.back()->pts == AV_NOPTS_VALUE){
+        reorder_queue.back()->pts = 0;
     }
 
     //compute the delay
@@ -365,7 +401,7 @@ bool StreamUnwrap::read_frame(void){
     }
     //check the time of the youngest packet...
     double delta = recv_time.tv_sec;
-    delta -= av_q2d(input_format_context->streams[reorder_queue.back().stream_index]->time_base) * reorder_queue.back().dts;
+    delta -= av_q2d(input_format_context->streams[reorder_queue.back()->stream_index]->time_base) * reorder_queue.back()->dts;
     if (delta > max_sync_offset || delta < -max_sync_offset){
         connect_time.tv_sec += delta;
         LWARN("Input stream has drifted " + std::to_string(delta) + "s from wall time on camera " + cfg.get_value("camera-id") + ", resyncing..." );
@@ -378,15 +414,15 @@ const struct timeval& StreamUnwrap::get_start_time(void){
     return connect_time;
 }
 
-bool StreamUnwrap::decode_packet(AVPacket &packet){
+bool StreamUnwrap::decode_packet(AVPacket *packet){
     int ret;
-    if (decode_ctx.find(packet.stream_index) == decode_ctx.end()){
-        if (!alloc_decode_context(packet.stream_index)){
+    if (decode_ctx.find(packet->stream_index) == decode_ctx.end()){
+        if (!alloc_decode_context(packet->stream_index)){
             return false;
         }
     }
 
-    ret = avcodec_send_packet(decode_ctx[packet.stream_index], &packet);
+    ret = avcodec_send_packet(decode_ctx[packet->stream_index], packet);
     if (ret == AVERROR(EAGAIN)){
         //probably should return ret instead of this as it's a soft error
         LWARN("Attempted to decode data without reading all frames");
@@ -445,8 +481,8 @@ bool StreamUnwrap::get_decoded_frame_int(int stream, AVFrame *frame){
     }
 }
 
-void StreamUnwrap::timestamp(const AVPacket &packet, struct timeval &time){
-    Util::compute_timestamp(get_start_time(), time, packet.pts, input_format_context->streams[packet.stream_index]->time_base);
+void StreamUnwrap::timestamp(const AVPacket *packet, struct timeval &time){
+    Util::compute_timestamp(get_start_time(), time, packet->pts, input_format_context->streams[packet->stream_index]->time_base);
 }
 
 void StreamUnwrap::timestamp(const AVFrame *frame, int stream_idx, struct timeval &time){
@@ -459,16 +495,16 @@ void StreamUnwrap::timestamp(const AVFrame *frame, int stream_idx, struct timeva
     Util::compute_timestamp(get_start_time(), time, frame->pts, input_format_context->streams[stream_idx]->time_base);
 }
 
-bool StreamUnwrap::is_audio(const AVPacket &packet){
-    return input_format_context->streams[packet.stream_index]->codecpar->codec_type == AVMEDIA_TYPE_AUDIO;
+bool StreamUnwrap::is_audio(const AVPacket *packet){
+    return input_format_context->streams[packet->stream_index]->codecpar->codec_type == AVMEDIA_TYPE_AUDIO;
 }
 
-bool StreamUnwrap::is_video(const AVPacket &packet){
-    return input_format_context->streams[packet.stream_index]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO;
+bool StreamUnwrap::is_video(const AVPacket *packet){
+    return input_format_context->streams[packet->stream_index]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO;
 }
 
-AVStream *StreamUnwrap::get_stream(const AVPacket &packet){
-    return get_stream(packet.stream_index);
+AVStream *StreamUnwrap::get_stream(const AVPacket *packet){
+    return get_stream(packet->stream_index);
 }
 
 AVStream *StreamUnwrap::get_stream(const int id){
@@ -506,32 +542,37 @@ AVCodecContext* StreamUnwrap::get_codec_context(AVStream *stream){
 }
 
 bool StreamUnwrap::charge_video_decoder(void){
-    decoded_packets.emplace_back();
-    if (!get_next_packet(decoded_packets.back())){
-        decoded_packets.pop_back();
-        return false;
-    }
-    AVPacket &pkt = decoded_packets.back();
-    if (decoded_video_frames.empty() && is_video(pkt)){
-
-        if (decode_packet(pkt)){
-            bool ret;
-            do {
-                AVFrame* next_frame = av_frame_alloc();
-                ret = get_decoded_frame_int(pkt.stream_index, next_frame);
-                if (ret){
-                    decoded_video_frames.push_back(next_frame);
-                    LDEBUG("Decoded frame");
-                    next_frame = NULL;
-                } else {
-                    av_frame_free(&next_frame);
-                }
-            } while (ret);
-        } else {
-            LWARN("Failed to decode video during charge");
+    do {
+        decoded_packets.emplace_back(av_packet_alloc());
+        if (!get_next_packet(decoded_packets.back())){
+            av_packet_free(&decoded_packets.back());
+            decoded_packets.pop_back();
             return false;
         }
-    }
+        AVPacket *pkt = decoded_packets.back();
+        if (decoded_video_frames.empty() && is_video(pkt)){
+            LDEBUG("Decoding packet");
+            if (decode_packet(pkt)){
+                LDEBUG("Decoded packet");
+                bool ret;
+                do {
+                    AVFrame* next_frame = av_frame_alloc();
+                    ret = get_decoded_frame_int(pkt->stream_index, next_frame);
+                    if (ret){
+                        decoded_video_frames.push_back(next_frame);
+                        LDEBUG("Decoded frame");
+                        next_frame = NULL;
+                        break;
+                    } else {
+                        av_frame_free(&next_frame);
+                    }
+                } while (ret);
+            } else {
+                LWARN("Failed to decode video during charge");
+                return false;
+            }
+        }
+    } while (decoded_video_frames.empty());
     return true;
 }
 
@@ -563,7 +604,7 @@ void StreamUnwrap::record_delay(const struct timeval &start, const struct timeva
     delay += (end.tv_usec - start.tv_usec)/1000000.0;
 
     //compute the mean time of the packets received
-    double duration = av_q2d(input_format_context->streams[reorder_queue.back().stream_index]->time_base) * reorder_queue.back().duration;
+    double duration = av_q2d(input_format_context->streams[reorder_queue.back()->stream_index]->time_base) * reorder_queue.back()->duration;
     duration /= input_format_context->nb_streams;
 
     if (timeshift_mean_delay != 0){
