@@ -36,11 +36,16 @@ StreamWriter::StreamWriter(Config& cfg) : cfg(cfg) {
     } else {
         segment_mode = SEGMENT_MPGTS;
     }
+    interleave_queue_depth = cfg.get_value_int("interleave-queue-depth");
+    if (interleave_queue_depth > 10000 || interleave_queue_depth < 0){
+        interleave_queue_depth = 4;
+    }
     video_stream_found = false;
     audio_stream_found = false;
     video_width = -1;
     video_height = -1;
     video_framerate = 0;
+    enc_pkt = av_packet_alloc();
 }
 
 bool StreamWriter::open(void){
@@ -124,6 +129,7 @@ bool StreamWriter::open_path(void){
 
 StreamWriter::~StreamWriter(){
     free_context();
+    av_packet_free(&enc_pkt);
 }
 
 long long StreamWriter::close(void){
@@ -150,19 +156,19 @@ long long StreamWriter::close(void){
     return pos;
 }
 
-bool StreamWriter::write(const AVPacket &packet, const AVStream *in_stream){
+bool StreamWriter::write(const AVPacket *packet, const AVStream *in_stream){
     if (!file_opened){
         return false;
     }
 
-    if (packet.stream_index >= static_cast<int>(stream_mapping.size()) ||
-        stream_mapping[packet.stream_index] < 0) {
+    if (packet->stream_index >= static_cast<int>(stream_mapping.size()) ||
+        stream_mapping[packet->stream_index] < 0) {
         return true;//we processed the stream we don't care about
     }
 
     AVStream *out_stream;
     PacketInterleavingBuf *pkt_buf = new PacketInterleavingBuf(packet);
-    pkt_buf->video_keyframe = in_stream->codecpar->codec_type == AVMEDIA_TYPE_VIDEO && packet.flags & AV_PKT_FLAG_KEY;
+    pkt_buf->video_keyframe = in_stream->codecpar->codec_type == AVMEDIA_TYPE_VIDEO && packet->flags & AV_PKT_FLAG_KEY;
 
     if (!pkt_buf->in || !pkt_buf->out){
         LERROR("Could not allocate new output packet for writing");
@@ -211,11 +217,11 @@ bool StreamWriter::write(const AVPacket &packet, const AVStream *in_stream){
 }
 
 bool StreamWriter::write(PacketInterleavingBuf *pkt_buf){
-
     if (keyframe_cbk && pkt_buf->video_keyframe){
-        keyframe_cbk(*(pkt_buf->in), *this);
+        keyframe_cbk(pkt_buf->in, *this);
     }
-    //LWARN("Writing: " + std::to_string(pkt_buf->dts0) + "(" + std::to_string(pkt_buf->out->dts) + ")");
+    //log_packet(output_format_context, pkt_buf->out, "Final Write");
+    write_counter += pkt_buf->out->size;//log this
     int ret = av_interleaved_write_frame(output_format_context, pkt_buf->out);
     if (ret < 0) {
         LERROR("Error muxing packet for camera " + cfg.get_value("camera-id"));
@@ -226,17 +232,17 @@ bool StreamWriter::write(PacketInterleavingBuf *pkt_buf){
     return true;
 }
 
-void StreamWriter::log_packet(const AVFormatContext *fmt_ctx, const AVPacket &pkt, const std::string &tag){
-    AVRational *time_base = &fmt_ctx->streams[pkt.stream_index]->time_base;
-    LINFO(tag + ": pts:" + std::string(av_ts2str(pkt.pts)) +
-          " pts_time:"+std::string(av_ts2timestr(pkt.pts, time_base))+
-          " dts:" + std::string(av_ts2str(pkt.dts)) +
-          " dts_time:"+ std::string(av_ts2timestr(pkt.dts, time_base)) +
-          " duration:" +std::string(av_ts2str(pkt.duration)) +
-          " duration_time:" + std::string(av_ts2timestr(pkt.duration, time_base))+
-          " stream_index:" + std::to_string(pkt.stream_index)
+void StreamWriter::log_packet(const AVFormatContext *fmt_ctx, const AVPacket *pkt, const std::string &tag){
+    AVRational *time_base = &fmt_ctx->streams[pkt->stream_index]->time_base;
+    LINFO(tag + ": pts:" + std::string(av_ts2str(pkt->pts)) +
+          " pts_time:"+std::string(av_ts2timestr(pkt->pts, time_base))+
+          " dts:" + std::string(av_ts2str(pkt->dts)) +
+          " dts_time:"+ std::string(av_ts2timestr(pkt->dts, time_base)) +
+          " duration:" +std::string(av_ts2str(pkt->duration)) +
+          " duration_time:" + std::string(av_ts2timestr(pkt->duration, time_base))+
+          " stream_index:" + std::to_string(pkt->stream_index)
     );
-    av_pkt_dump_log2(NULL, 30, &pkt, 0, fmt_ctx->streams[pkt.stream_index]);
+    av_pkt_dump_log2(NULL, 30, pkt, 0, fmt_ctx->streams[pkt->stream_index]);
 }
 
 long long StreamWriter::change_path(const std::string &new_path /* = "" */){
@@ -371,8 +377,13 @@ bool StreamWriter::add_encoded_stream(const AVStream *in_stream, const AVCodecCo
                 return false;
             }
             encode_ctx[out_stream->index]->sample_rate = dec_ctx->sample_rate;
+            //this changes in FFMpeg 5.1
+#if LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(57, 28, 100)
+            encode_ctx[out_stream->index]->ch_layout = dec_ctx->ch_layout;
+#else
             encode_ctx[out_stream->index]->channel_layout = dec_ctx->channel_layout;
             encode_ctx[out_stream->index]->channels = av_get_channel_layout_nb_channels(encode_ctx[out_stream->index]->channel_layout);
+#endif
             encode_ctx[out_stream->index]->sample_fmt = encoder->sample_fmts[0];
             encode_ctx[out_stream->index]->time_base = (AVRational){1, encode_ctx[out_stream->index]->sample_rate};
 
@@ -450,7 +461,11 @@ bool StreamWriter::add_encoded_stream(const AVStream *in_stream, const AVCodecCo
             encode_ctx[out_stream->index]->rc_max_rate = encode_ctx[out_stream->index]->bit_rate + (encode_ctx[out_stream->index]->bit_rate/10);
             encode_ctx[out_stream->index]->rc_min_rate = encode_ctx[out_stream->index]->bit_rate - (encode_ctx[out_stream->index]->bit_rate/10);
 
-            encode_ctx[out_stream->index]->max_b_frames = 16;
+            int bframe_max = cfg.get_value_int("video-encode-b-frame-max");
+            if (bframe_max < 0){
+                bframe_max = 16;
+            }
+            encode_ctx[out_stream->index]->max_b_frames = bframe_max;
 
             LINFO("Selected video encode bitrate: " + std::to_string(encode_ctx[out_stream->index]->bit_rate/1000) + "kbps");
             //apply encode video options
@@ -458,8 +473,10 @@ bool StreamWriter::add_encoded_stream(const AVStream *in_stream, const AVCodecCo
             av_dict_copy(&opts, vopts, 0);
             av_dict_free(&vopts);
 
+            //check if v4l2 was selected
+            bool using_v4l2 = std::string(encode_ctx[out_stream->index]->codec->name).find("_v4l2m2m") != std::string::npos;
             //connect encoder, try VA-API
-            if (!encode_ctx[out_stream->index]->hw_device_ctx &&
+            if (!using_v4l2 && !encode_ctx[out_stream->index]->hw_device_ctx &&
                 (cfg.get_value("video-encode-method") == "auto" || cfg.get_value("video-encode-method") == "vaapi")){
                 encode_ctx[out_stream->index]->hw_device_ctx = gcff_util.get_vaapi_ctx(encode_ctx[out_stream->index]->codec_id, encode_ctx[out_stream->index]->profile,
                                                                                        encode_ctx[out_stream->index]->width, encode_ctx[out_stream->index]->height);
@@ -475,6 +492,8 @@ bool StreamWriter::add_encoded_stream(const AVStream *in_stream, const AVCodecCo
                         LWARN("Wrong pixel format provided for initilization");
                     }
                 }
+            } else if (using_v4l2 && dec_ctx->pix_fmt == AV_PIX_FMT_DRM_PRIME){
+                encode_ctx[out_stream->index]->pix_fmt = AV_PIX_FMT_DRM_PRIME;//if using V4L2 and the source was in DRM, keep using DRM
             }
             //if it fails we get the SW encoder
         } else {
@@ -523,17 +542,13 @@ bool StreamWriter::add_encoded_stream(const AVStream *in_stream, const AVCodecCo
 
 bool StreamWriter::write(const AVFrame *frame, const AVStream *in_stream){
     int ret = 0;
-    AVPacket enc_pkt;
-    av_init_packet(&enc_pkt);
-    enc_pkt.data = NULL;
-    enc_pkt.size = 0;
     ret = avcodec_send_frame(encode_ctx[stream_mapping[in_stream->index]], frame);
     if (ret < 0){
         LWARN("Error during encoding. Error code: " +  std::string(av_err2str(ret)));
         return false;
     }
     while (1) {
-        ret = avcodec_receive_packet(encode_ctx[stream_mapping[in_stream->index]], &enc_pkt);
+        ret = avcodec_receive_packet(encode_ctx[stream_mapping[in_stream->index]], enc_pkt);
         if (ret){
             if (ret == AVERROR(EAGAIN)){
                 break;
@@ -542,9 +557,9 @@ bool StreamWriter::write(const AVFrame *frame, const AVStream *in_stream){
             break;
         }
 
-        enc_pkt.stream_index = in_stream->index;//revert stream index because write will adjust this
+        enc_pkt->stream_index = in_stream->index;//revert stream index because write will adjust this
         bool write_ret = write(enc_pkt, in_stream);
-        av_packet_unref(&enc_pkt);
+        av_packet_unref(enc_pkt);
         if (!write_ret){
             return false;
         }
@@ -615,7 +630,7 @@ long long StreamWriter::get_init_len(void) const {
     return init_len;
 }
 
-void StreamWriter::set_keyframe_callback(std::function<void(const AVPacket &pkt, StreamWriter &out)> cbk){
+void StreamWriter::set_keyframe_callback(std::function<void(const AVPacket *pkt, StreamWriter &out)> cbk){
     keyframe_cbk = cbk;
 }
 
@@ -671,14 +686,36 @@ void StreamWriter::interleave(PacketInterleavingBuf *buf){
                                      output_format_context->streams[0]->time_base, AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
     }
     interleaved_dts[buf->out->stream_index] = buf->dts0;
-    //find the first buf where dts0 is > than ours, and insert just before that
+    //find the first buf where dts0 is > than ours, and insert jus before that
     for (auto ibuf = interleaving_buf.begin(); ibuf != interleaving_buf.end(); ibuf++){
         if ((*ibuf)->dts0 > buf->dts0){
             interleaving_buf.insert(ibuf, buf);
             return;
         }
     }
-    interleaving_buf.push_back(buf);
+    interleaving_buf.insert(interleaving_buf.end(), buf);
+}
+
+bool StreamWriter::fixup_duration_interleave(std::list<PacketInterleavingBuf*>::iterator itr){
+    auto in = *itr;
+    auto end_dts = in->out->dts + in->out->duration;
+    for (itr++; itr != interleaving_buf.end(); itr++){
+        if ((*itr)->out->stream_index == in->out->stream_index){
+            //found the previous packet
+            if (end_dts > (*itr)->out->dts){//our current DTS puts us past the next packet start
+                //log_packet(output_format_context, in->out, "Pre-fix");
+                auto target_duration = (*itr)->out->dts - in->out->dts;
+                if (target_duration > 0){
+                    in->out->duration = target_duration;
+                }
+                //log_packet(output_format_context, in->out, "Post-fix");
+                return true;
+            } else {
+                return false;
+            }
+        }
+    }
+    return false;
 }
 
 bool StreamWriter::write_interleaved(void){
@@ -705,7 +742,26 @@ bool StreamWriter::write_interleaved(void){
 
     bool ret = true;
     for (auto ibuf = interleaving_buf.begin(); ibuf != interleaving_buf.end();){
-        if ((*ibuf)->dts0 <= last_dts){
+        if ((*ibuf)->dts0 <= last_dts && interleaving_buf.size() > interleave_queue_depth){
+            //fixup the PTS and DTS
+            try {
+                long track_dts = track_len_dts.at((*ibuf)->out->stream_index);//throws and we don't check
+                if ((*ibuf)->out->dts > track_dts){
+                    long diff = (*ibuf)->out->dts - track_dts;
+                    if (diff < (*ibuf)->out->duration){
+                        //difference is less than a packet, shift DTS and PTS
+                        (*ibuf)->out->dts -= diff;
+                        (*ibuf)->out->pts -= diff;
+                    } else {
+                        LDEBUG("Excessive DTS gap (" + std::to_string(diff) + "), possible dropped packet");
+                    }
+                }
+            } catch (std::out_of_range &e){
+                //track dts is not set (first run)
+            }
+            track_len_dts[(*ibuf)->out->stream_index] = (*ibuf)->out->dts + (*ibuf)->out->duration;
+
+            fixup_duration_interleave(ibuf);
             ret &= write(*ibuf);
             ibuf = interleaving_buf.erase(ibuf);
         } else {
@@ -913,7 +969,7 @@ bool StreamWriter::get_video_format(const AVFrame *frame, AVPixelFormat &pix_fmt
     const AVCodec *vencoder = get_vencoder(frame->width, frame->height, codec_id, codec_profile);
     if (!vencoder || !vencoder->pix_fmts){
         LWARN("Video Encoder has no pixel formats?");
-        pix_fmt = AV_PIX_FMT_NONE;
+        pix_fmt = static_cast<AVPixelFormat>(frame->format);//I guess the encoder takes anything...see what this does
         return false;
     }
     //FIXME: This is overly restrictive, our API should return all pix_fmts
@@ -931,6 +987,11 @@ const AVCodec* StreamWriter::get_vencoder(int width, int height, AVCodecID &code
             gcff_util.have_vaapi(codec_id, codec_profile, width, height)){
             encoder = avcodec_find_encoder_by_name("hevc_vaapi");
         }
+        if (!encoder && ((cfg.get_value("video-encode-method") == "auto" || cfg.get_value("video-encode-method") == "vaapi") &&
+                         gcff_util.have_v4l2(codec_id, codec_profile, width, height))){
+            encoder = avcodec_find_encoder_by_name("hevc_v4l2m2m");
+        }
+
         if (!encoder){
             encoder = avcodec_find_encoder(AV_CODEC_ID_HEVC);
         }
@@ -943,6 +1004,11 @@ const AVCodec* StreamWriter::get_vencoder(int width, int height, AVCodecID &code
             gcff_util.have_vaapi(AV_CODEC_ID_H264, FF_PROFILE_H264_MAIN, width, height)){
             encoder = avcodec_find_encoder_by_name("h264_vaapi");
         }
+        if (!encoder && ((cfg.get_value("video-encode-method") == "auto" || cfg.get_value("video-encode-method") == "v4l2") &&
+            gcff_util.have_v4l2(AV_CODEC_ID_H264, FF_PROFILE_H264_MAIN, width, height))){
+            encoder = avcodec_find_encoder_by_name("h264_v4l2m2m");
+        }
+
         if (!encoder){
             encoder = avcodec_find_encoder(AV_CODEC_ID_H264);
         }
@@ -956,9 +1022,15 @@ const AVCodec* StreamWriter::get_vencoder(int width, int height) const {
     return get_vencoder(width, height, codec_id, codec_profile);
 }
 
-PacketInterleavingBuf::PacketInterleavingBuf(const AVPacket &pkt){
-    in = av_packet_clone(&pkt);
-    out = av_packet_clone(&pkt);
+unsigned int StreamWriter::get_bytes_written(void){
+    auto bytes = write_counter;
+    write_counter = 0;
+    return bytes;
+}
+
+PacketInterleavingBuf::PacketInterleavingBuf(const AVPacket *pkt){
+    in = av_packet_clone(pkt);
+    out = av_packet_clone(pkt);
     video_keyframe = false;
 }
 
