@@ -568,7 +568,7 @@ class CameraState {
         }
 
         for (var i = 0; i < this.jsonData.gaps.length; i++){
-            if (this.jsonData.gaps[i].actual_start_ts < targetTime){
+            if (this.jsonData.gaps[i].actual_start_ts < (targetTime - totalGaps)){
                 totalGaps += this.jsonData.gaps[i].len;
             } else {
                 //only works if we guarentee gaps is in order, which it should be
@@ -701,6 +701,8 @@ function setDefaultOpt(id){
     } else {
         textVal.value = settings[select.value].value;
     }
+    let descBox = select.parentElement.parentElement.querySelector(".setting_desc");
+    descBox.innerHTML = settings[select.value].desc;
 }
 
 function clearSystemCheckbox(caller, id){
